"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster, SpecificUser
from typing import TYPE_CHECKING, TypedDict
from .chat import TextEmoteMessage

if TYPE_CHECKING:
    from typing import Optional, List, Literal


# Reward
class Emote(TypedDict):
    """
    Represents an emote with its ID and availability times.

    Attributes
    ----------
    id: str
        The ID of the emote.
    begin: int
        The start time of the emotes availability, in Unix timestamp format.
    end: int
        The end time of the emotes availability, in Unix timestamp format.
    """
    id: str
    begin: int
    end: int


class UnlockedEmote(TypedDict):
    """
    Represents an emote that has been unlocked.

    Attributes
    ----------
    id: str
        The ID of the unlocked emote.
    name: str
        The name of the unlocked emote.
    """
    id: str
    name: str


class AutomaticReward(TypedDict):
    """
    Represents an automatic reward with its type and details.

    Attributes
    ----------
    type: str
        The type of event. Possible values include:

        - single_message_bypass_sub_mode
        - send_highlighted_message
        - random_sub_emote_unlock
        - chosen_sub_emote_unlock
        - chosen_modified_sub_emote_unlock
        - message_effect
        - gigantify_an_emote
        - celebration
    cost: int
        The cost of the reward in channel points or other currency.
    unlocked_emote: Optional[UnlockedEmote]
        Details of an emote unlocked by the reward, if applicable.
    """
    type: Literal['single_message_bypass_sub_mode',
                  'send_highlighted_message',
                  'random_sub_emote_unlock',
                  'chosen_sub_emote_unlock',
                  'chosen_modified_sub_emote_unlock',
                  'message_effect',
                  'gigantify_an_emote',
                  'celebration']
    cost: int
    unlocked_emote: Optional[UnlockedEmote]


class AutomaticRewardRedemptionAddEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where an automatic reward is redeemed.

    Attributes
    ----------
    id: str
        The ID of the reward redemption event.
    reward: AutomaticReward
        The details of the automatic reward redeemed.
    message: TextEmoteMessage
        The message associated with the reward redemption.
    user_input: Optional[str]
        Additional user input, if any, provided during the redemption.
    redeemed_at: str
        The timestamp when the reward was redeemed, in ISO 8601 format.
    """
    id: str
    reward: AutomaticReward
    message: TextEmoteMessage
    user_input: Optional[str]
    redeemed_at: str


class Image(TypedDict):
    """
    Represents an image with multiple resolution options.

    Attributes
    ----------
    url_1x: Optional[str]
        URL of the image at 1x resolution.
    url_2x: Optional[str]
        URL of the image at 2x resolution.
    url_4x: Optional[str]
        URL of the image at 4x resolution.
    """
    url_1x: Optional[str]
    url_2x: Optional[str]
    url_4x: Optional[str]


class PointRewardEvent(SpecificBroadcaster):
    """
    Represents custom point reward event.

    Attributes
    ----------
    id: str
        The ID of the reward.
    is_enabled: bool
        Whether the reward is currently enabled.
    is_paused: bool
        Whether the reward is currently paused.
    is_in_stock: bool
        Whether the reward is in stock.
    title: str
        The title of the reward.
    cost: int
        The cost of the reward in channel points or other currency.
    prompt: str
        The prompt or description for the reward.
    is_user_input_required: bool
        Whether user input is required to redeem the reward.
    should_redemptions_skip_request_queue: bool
        Whether redemptions should skip the request queue.
    max_per_stream: Optional[int]
        The maximum number of redemptions allowed per stream, if applicable.
    max_per_user_per_stream: Optional[int]
        The maximum number of redemptions allowed per user per stream, if applicable.
    background_color: Optional[str]
        The background color of the reward's display.
    image: Optional[Image]
        The image associated with the reward.
    default_image: Optional[Image]
        The default image to display for the reward.
    global_cooldown: Optional[int]
        The global cooldown for the reward, in seconds.
    cooldown_expires_at: Optional[str]
        The timestamp when the cooldown expires, in ISO 8601 format.
    redemptions_redeemed_current_stream: Optional[int]
        The number of redemptions redeemed in the current stream.
    """
    id: str
    is_enabled: bool
    is_paused: bool
    is_in_stock: bool
    title: str
    cost: int
    prompt: str
    is_user_input_required: bool
    should_redemptions_skip_request_queue: bool
    max_per_stream: Optional[int]
    max_per_user_per_stream: Optional[int]
    background_color: Optional[str]
    image: Optional[Image]
    default_image: Optional[Image]
    global_cooldown: Optional[int]
    cooldown_expires_at: Optional[str]
    redemptions_redeemed_current_stream: Optional[int]


class RewardInfo(TypedDict):
    """
    Represents basic information about a reward.

    Attributes
    ----------
    id: str
        The ID of the reward.
    title: str
        The title of the reward.
    cost: int
        The cost of the reward in channel points or other currency.
    """
    id: str
    title: str
    cost: int


class RewardRedemptionEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a reward redemption occurs.

    Attributes
    ----------
    id: str
        The ID of the reward redemption event.
    user_input: str
        The user input provided during the reward redemption.
    status: Literal['unknown', 'unfulfilled', 'fulfilled', 'canceled']
        The current status of the reward redemption.
    reward: RewardInfo
        Basic information about the reward redeemed.
    redeemed_at: str
        The timestamp when the reward was redeemed, in ISO 8601 format.
    """
    id: str
    user_input: str
    status: Literal['unknown', 'unfulfilled', 'fulfilled', 'canceled']
    reward: RewardInfo
    redeemed_at: str


# Poll
class Choice(TypedDict):
    """
    Represents a choice in a poll.

    Attributes
    ----------
    id: str
        The ID of the choice.
    title: str
        The title of the choice.
    bits_votes: int
        The number of votes cast with bits for this choice.
    channel_points_votes: int
        The number of votes cast with channel points for this choice.
    """
    id: str
    title: str
    bits_votes: int
    channel_points_votes: int


class Voting(TypedDict):
    """
    Represents the voting configuration for a poll.

    Attributes
    ----------
    is_enabled: bool
        Whether voting is enabled for the poll.
    amount_per_vote: int
        The amount of currency required per vote.
    """
    is_enabled: bool
    amount_per_vote: int


class PollBase(SpecificBroadcaster):
    """
    Represents the base attributes for a poll.

    Attributes
    ----------
    id: str
        The ID of the poll.
    title: str
        The title of the poll.
    choices: List[Choice]
        A list of choices available in the poll.
    bits_voting: Voting
        The voting configuration for bits.
    channel_points_voting: Voting
        The voting configuration for channel points.
    started_at: str
        The timestamp when the poll started, in ISO 8601 format.
    """
    id: str
    title: str
    choices: List[Choice]
    bits_voting: Voting
    channel_points_voting: Voting
    started_at: str


class PollBeginEvent(PollBase):
    """
    Represents an event where a poll begins.

    Attributes
    ----------
    ends_at: str
        The timestamp when the poll ends, in ISO 8601 format.
    """
    ends_at: str


class PollProgressEvent(PollBase):
    """
    Represents an event where a poll is in progress.

    Attributes
    ----------
    ends_at: str
        The timestamp when the poll ends, in ISO 8601 format.
    """
    ends_at: str


class PollEndEvent(PollBase):
    """
    Represents an event where a poll ends.

    Attributes
    ----------
    status: Literal['completed', 'archived', 'terminated']
        The status of the poll at the end.
    ended_at: str
        The timestamp when the poll ended, in ISO 8601 format.
    """
    status: Literal['completed', 'archived', 'terminated']
    ended_at: str


# Prediction
class TopPredictor(SpecificUser):
    """
    Represents a top predictor in a prediction.

    Attributes
    ----------
    channel_points_used: int
        The amount of channel points used by the top predictor.
    channel_points_won: Optional[int]
        The amount of channel points won by the top predictor, if applicable.
    """
    channel_points_used: int
    channel_points_won: Optional[int]


class Outcome(TypedDict):
    """
    Represents an outcome in a prediction.

    Attributes
    ----------
    id: str
        The ID of the outcome.
    title: str
        The title of the outcome.
    color: Literal['BLUE', 'PINK']
        The color associated with the outcome.
    users: int
        The number of users who predicted this outcome.
    channel_points: int
        The total channel points associated with this outcome.
    top_predictors: Optional[List[TopPredictor]]
        A list of top predictors for this outcome, if applicable.
    """
    id: str
    title: str
    color: Literal['BLUE', 'PINK']
    users: int
    channel_points: int
    top_predictors: Optional[List[TopPredictor]]


class PredictionBase(SpecificBroadcaster):
    """
    Represents the base attributes for a prediction.

    Attributes
    ----------
    id: str
        The ID of the prediction.
    title: str
        The title of the prediction.
    outcomes: List[Outcome]
        A list of outcomes for the prediction.
    started_at: str
        The timestamp when the prediction started, in ISO 8601 format.
    """
    id: str
    title: str
    outcomes: List[Outcome]
    started_at: str


class PredictionBeginEvent(PredictionBase):
    """
    Represents an event where a prediction begins.

    Attributes
    ----------
    locks_at: str
        The timestamp when the prediction locks, in ISO 8601 format.
    """
    locks_at: str


class PredictionProgressEvent(PredictionBase):
    """
    Represents an event where a prediction is in progress.

    Attributes
    ----------
    locks_at: str
        The timestamp when the prediction locks, in ISO 8601 format.
    """
    locks_at: str


class PredictionLockEvent(PredictionBase):
    """
    Represents an event where a prediction locks.

    Attributes
    ----------
    locked_at: str
        The timestamp when the prediction was locked, in ISO 8601 format.
    """
    locked_at: str


class PredictionEndEvent(PredictionBase):
    """
    Represents an event where a prediction ends.

    Attributes
    ----------
    winning_outcome_id: str
        The ID of the winning outcome.
    status: Literal['resolved', 'canceled']
        The status of the prediction at the end.
    ended_at: str
        The timestamp when the prediction ended, in ISO 8601 format.
    """
    winning_outcome_id: str
    status: Literal['resolved', 'canceled']
    ended_at: str


# HypeTrain
class Contribution(SpecificUser):
    """
    Represents a contribution to a Hype Train.

    Attributes
    ----------
    type: Literal['bits', 'subscription']
        The type of contribution.
    total: int
        The total amount of the contribution.
    """
    type: Literal['bits', 'subscription']
    total: int


class HypeTrainEvent(SpecificBroadcaster):
    """
    Represents a Hype Train event.

    Attributes
    ----------
    id: str
        The ID of the Hype Train event.
    total: int
        The total amount of contributions to the Hype Train.
    progress: int
        The current progress of the Hype Train.
    goal: int
        The goal amount for the Hype Train.
    top_contributions: List[Contribution]
        A list of top contributions to the Hype Train.
    last_contribution: Contribution
        The most recent contribution to the Hype Train.
    started_at: str
        The timestamp when the Hype Train started, in ISO 8601 format.
    expires_at: str
        The timestamp when the Hype Train expires, in ISO 8601 format.
    """
    id: str
    total: int
    progress: int
    goal: int
    top_contributions: List[Contribution]
    last_contribution: Contribution
    started_at: str
    expires_at: str


class HypeTrainEndEvent(SpecificBroadcaster):
    """
    Represents an event where a Hype Train ends.

    Attributes
    ----------
    id: str
        The ID of the Hype Train event.
    level: int
        The level reached by the Hype Train.
    total: int
        The total amount of contributions to the Hype Train.
    top_contributions: List[Contribution]
        A list of top contributions to the Hype Train.
    started_at: str
        The timestamp when the Hype Train started, in ISO 8601 format.
    ended_at: str
        The timestamp when the Hype Train ended, in ISO 8601 format.
    cooldown_ends_at: str
        The timestamp when the cooldown period ends, in ISO 8601 format.
    """
    id: str
    level: int
    total: int
    top_contributions: List[Contribution]
    started_at: str
    ended_at: str
    cooldown_ends_at: str



"""
twitch.types

Typing for Twitch.

:copyright: (c) 2024-present Snifo
:license: MIT, see LICENSE for more details.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import TypeVar
    T = TypeVar('T')


# Most of the data in the Helix API is wrapped with the 'data' keyword.
class Data(TypedDict):
    data: T


pagination = TypedDict('pagination', {'cursor': str})


class PData(Data):
    pagination: pagination


date_range = TypedDict('date_range', {'started_at': str, 'ended_at': str})


class TData(PData):
    total: int


# EventSub Subscription
class TTMData(TData):
    total_cost: int
    max_total_cost: int


# Subscription
class TPData(PData):
    points: int
    total: int


# Chat Emote
class Edata(Data):
    template: str


class PEdata(PData):
    template: str


class DTData(Data):
    date_range: date_range
    total: int



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict
from .users import SpecificUser

if TYPE_CHECKING:
    from typing import List, Literal


# Leaderboard
class Leaderboard(SpecificUser):
    """
    Represents a leaderboard entry for a user.

    Attributes
    ----------
    rank: int
        The rank of the user in the leaderboard.
    score: int
        The score of the user in the leaderboard.
    """
    rank: int
    score: int


# Cheermote
CheermoteImages = TypedDict('CheermoteImages', {'1': str, '1.5': str, '2': str, '3': str, '4': str})


class ImagesType(TypedDict):
    """
    Represents different types of cheermote images.

    Attributes
    ----------
    animated: CheermoteImages
        A dictionary containing URLs of cheermote images for different bit amounts in animated form.
    static: CheermoteImages
        A dictionary containing URLs of cheermote images for different bit amounts in static form.

    The `CheermoteImages` dictionary contains keys for various bit amounts, including:
    - '1': URL for 1 bit
    - '1.5': URL for 1.5 bits
    - '2': URL for 2 bits
    - '3': URL for 3 bits
    - '4': URL for 4 bits
    """
    animated: CheermoteImages
    static: CheermoteImages


class TierImages(TypedDict):
    """
    Represents cheermote images categorized by theme mode.

    Attributes
    ----------
    dark: ImagesType
        A dictionary containing URLs of cheermote images categorized by type (animated or static) for dark mode.
    light: ImagesType
        A dictionary containing URLs of cheermote images categorized by type (animated or static) for light mode.
    """
    dark: ImagesType
    light: ImagesType


class CheermoteTier(TypedDict):
    """
    Represents a tier of cheermote with specific attributes.

    Attributes
    ----------
    id: str
        The unique identifier of the cheermote tier.
    min_bits: Literal[1, 100, 500, 1000, 5000, 10000, 100000]
        The minimum number of bits required to unlock this cheermote tier.
    color: str
        The color associated with the cheermote tier.
    images: TierImages
        A dictionary containing cheermote images categorized by theme mode and type.
    can_cheer: bool
        Whether the cheermote tier can be used for cheering.
    show_in_bits_card: bool
        Whether the cheermote tier is shown in the bits card.
    """
    id: str
    min_bits: Literal[1, 100, 500, 1000, 5000, 10000, 100000]
    color: str
    images: TierImages
    can_cheer: bool
    show_in_bits_card: bool


class Cheermote(TypedDict):
    """
    Represents a cheermote with its details and tiers.

    Attributes
    ----------
    prefix: str
        The prefix used to represent the cheermote.
    tiers: List[CheermoteTier]
        A list of cheermote tiers available.
    type: Literal['global_first_party', 'global_third_party', 'channel_custom', 'display_only', 'sponsored']
        The type of cheermote.
    order: int
        The order of the cheermote.
    last_updated: str
        The timestamp when the cheermote was last updated.
    is_charitable: bool
        Whether the cheermote is used for charitable purposes.
    """
    prefix: str
    tiers: List[CheermoteTier]
    type: Literal['global_first_party', 'global_third_party', 'channel_custom', 'display_only', 'sponsored']
    order: int
    last_updated: str
    is_charitable: bool



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, overload
import twitch

if TYPE_CHECKING:
    from twitch.user import User, Broadcaster
    from typing import Optional, List


import logging
_logger = logging.getLogger(__name__)

__all__ = ('Bot',)


class Bot(twitch.Client):
    """
    A class that extends the functionality of the `Client` class.

    This extension adds capabilities for managing multiple users and their tokens,
    allowing the bot to perform actions on behalf of registered users for more flexibility.

    Parameters
    ----------
    client_id: str
        The client ID used to authenticate with the Twitch API.
    client_secret: Optional[str]
        The client secret used for authentication. Default is None.
    **options:
        Additional configuration options:

        - cli: [bool][bool]

            Flag indicating if CLI mode is enabled.

            **Default**: `False`

            ___

        - cli_port: [int][]

            The port number used for CLI mode.

            **Default**: `8080`
            ___

        - socket_debug: [False][bool]

            Flag indicating if raw WebSocket messages should be dispatched for debugging purposes.
            If enabled, raw WebSocket messages are dispatched to the on_socket_raw_receive.

            **Default**: `False`
            ___

        - proxy: [Optional][typing.Optional][[str][str]]

            The proxy URL to use for HTTP requests.

            **Default**: `None`
            ___

        - proxy_auth: [Optional][typing.Optional][aiohttp.BasicAuth]

            Authentication details for the proxy, if required.

            **Default**: `None`
    """
    def __init__(self, client_id: str, client_secret: Optional[str] = None, **options) -> None:
        super().__init__(client_id=client_id, client_secret=client_secret, **options)

    def get_broadcasters(self) -> List[Broadcaster]:
        """
        Retrieve a list of registered broadcasters.

        This method fetches and returns a list of all broadcasters that have been registered
        and are managed by the connection.

        Returns
        -------
        List[Broadcaster]
            A list of Broadcaster objects representing all registered broadcasters.
        """
        return self._connection.get_broadcasters()

    async def get_broadcaster(self, user: User) -> Broadcaster:
        """
        Retrieve a registered broadcaster.

        Parameters
        ----------
        user: User
            The user whose broadcaster details are being requested.

        Returns
        -------
        Broadcaster
            The broadcaster object corresponding to the user.
        """
        return await self._connection.get_broadcaster(user.id)

    async def get_broadcaster_by_id(self, __id: str, /) -> Broadcaster:
        """
        Retrieve a registered broadcaster using their ID.

        Parameters
        ----------
        __id: str
            The ID of the broadcaster.

        Returns
        -------
        Broadcaster
            The broadcaster object corresponding to the ID.
        """
        return await self._connection.get_broadcaster(__id)

    def is_registered(self, user: User) -> bool:
        """
        Check if a broadcaster is registered with the bot.

        Parameters
        ----------
        user: User
            The user to check registration status for.

        Returns
        -------
        bool
            True if the user is registered, False otherwise.
        """
        return self._connection.is_registered(user.id)

    @overload
    async def register_user(self, *, access_token: str, refresh_token: None = None) -> Broadcaster:
        ...

    @overload
    async def register_user(self, *, refresh_token: str, access_token: None = None) -> Broadcaster:
        ...

    @overload
    async def register_user(self, *, access_token: str, refresh_token: str) -> Broadcaster:
        ...

    async def register_user(self, *,
                            access_token: Optional[str] = None,
                            refresh_token: Optional[str] = None) -> Broadcaster:
        """
        Register a broadcaster with an access token, refresh token, or both.

        !!! Warning
            If you are planning to use only a refresh token, a client secret is required.

        Parameters
        ----------
        access_token: Optional[str]
            Access token for the user.
        refresh_token: Optional[str]
            Refresh token for the user.

        Returns
        -------
        Broadcaster
            The broadcaster object corresponding to the registered user.
        """
        if access_token is None and refresh_token:
            if self.client_secret is None:
                raise TypeError('Missing client secret.')

        return await self._connection.register_user(access_token, refresh_token)

    async def unregister_user(self, user: User) -> None:
        """
        Unregister a broadcaster from the bot.

        Parameters
        ----------
        user: User
            The user to remove from registration.
        """
        await self._connection.remove_user(user.id)



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import Broadcaster, SpecificUser
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import List, Literal, Dict, Optional, NotRequired


# Channel
class CCL(TypedDict):
    """
    Represents a Content Classification Label (CCL).

    Attributes
    ----------
    id: Literal['DrugsIntoxication', 'SexualThemes', 'ViolentGraphic', 'Gambling', 'ProfanityVulgarity']
        The type of content classification.
    is_enabled: bool
        Indicates if the classification is enabled.
    """
    id: Literal['DrugsIntoxication', 'SexualThemes', 'ViolentGraphic', 'Gambling', 'ProfanityVulgarity']
    is_enabled: bool


class ChannelInfo(Broadcaster):
    """
    Represents channel information.

    Attributes
    ----------
    broadcaster_language: str
        The language of the broadcaster.
    game_id: str
        The ID of the current game being played on the channel.
    game_name: str
        The name of the current game being played on the channel.
    title: str
        The title of the broadcast.
    delay: int
        The delay of the broadcast in seconds.
    tags: List[str]
        A list of tags associated with the broadcast.
    content_classification_labels: List[CCL]
        A list of content classification labels for the broadcast.
    is_branded_content: bool
        Indicates if the content is branded.
    """
    broadcaster_language: str
    game_id: str
    game_name: str
    title: str
    delay: int
    tags: List[str]
    content_classification_labels: List[CCL]
    is_branded_content: bool


class Editor(TypedDict):
    """
    Represents an editor of the channel.

    Attributes
    ----------
    user_id: str
        The ID of the editor.
    user_name: str
        The name of the editor.
    created_at: str
        The timestamp when the editor was added.
    """
    user_id: str
    user_name: str
    created_at: str


class Follows(Broadcaster):
    """
    Represents a follow relationship.

    Attributes
    ----------
    followed_at: str
        The timestamp when the channel was followed.
    """
    followed_at: str


class Follower(SpecificUser):
    """
    Represents a follower.

    Attributes
    ----------
    followed_at: str
        The timestamp when the user followed the channel.
    """
    followed_at: str


# Subscription
class Subscription(Broadcaster, SpecificUser):
    """
    Represents a subscription.

    Attributes
    ----------
    gifter_id: str
        The ID of the user who gifted the subscription.
    gifter_login: str
        The login name of the user who gifted the subscription.
    gifter_name: str
        The display name of the user who gifted the subscription.
    is_gift: bool
        Indicates if the subscription was a gift.
    plan_name: str
        The name of the subscription plan.
    tier: Literal['1000', '2000', '3000']
        The tier of the subscription.
    """
    gifter_id: str
    gifter_login: str
    gifter_name: str
    is_gift: bool
    plan_name: str
    tier: Literal['1000', '2000', '3000']


class SubscriptionCheck(Broadcaster, total=False):
    """
    Represents check subscription details.

    Attributes
    ----------
    gifter_id: NotRequired[str]
        The ID of the user who gifted the subscription.
    gifter_login: NotRequired[str]
        The login name of the user who gifted the subscription.
    gifter_name: NotRequired[str]
        The display name of the user who gifted the subscription.
    is_gift: bool
        Indicates if the subscription was a gift.
    tier: Literal['1000', '2000', '3000']
        The tier of the subscription.
    """
    gifter_id: NotRequired[str]
    gifter_login: NotRequired[str]
    gifter_name: NotRequired[str]
    is_gift: bool
    tier: Literal['1000', '2000', '3000']


# Clip
class ClipEdit(TypedDict):
    """
    Represents a clip edit.

    Attributes
    ----------
    edit_url: str
        The URL to edit the clip.
    id: str
        The ID of the clip.
    """
    edit_url: str
    id: str


class Clip(TypedDict):
    """
    Represents a clip.

    Attributes
    ----------
    id: str
        The ID of the clip.
    url: str
        The URL of the clip.
    title: str
        The title of the clip.
    game_id: str
        The ID of the game being played in the clip.
    duration: float
        The duration of the clip in seconds.
    language: str
        The language of the clip.
    video_id: str
        The ID of the video from which the clip was created.
    embed_url: str
        The URL to embed the clip.
    created_at: str
        The timestamp when the clip was created.
    creator_id: str
        The ID of the creator of the clip.
    view_count: int
        The number of views the clip has.
    vod_offset: Optional[int]
        The offset in the VOD where the clip starts.
    creator_name: str
        The display name of the creator of the clip.
    thumbnail_url: str
        The URL of the clip's thumbnail.
    broadcaster_id: str
        The ID of the broadcaster of the clip.
    broadcaster_name: str
        The display name of the broadcaster of the clip.
    """
    id: str
    url: str
    title: str
    game_id: str
    duration: float
    language: str
    video_id: str
    embed_url: str
    created_at: str
    creator_id: str
    view_count: int
    vod_offset: Optional[int]
    creator_name: str
    thumbnail_url: str
    broadcaster_id: str
    broadcaster_name: str


# Video
class VideoMutedSegments(TypedDict):
    """
    Represents muted segments in a video.

    Attributes
    ----------
    duration: int
        The duration of the muted segment in seconds.
    offset: int
        The offset in the video where the muted segment starts.
    """
    duration: int
    offset: int


class Video(SpecificUser):
    """
    Represents a video.

    Attributes
    ----------
    id: str
        The ID of the video.
    url: str
        The URL of the video.
    type: Literal['archive', 'highlight', 'upload']
        The type of video.
    title: str
        The title of the video.
    duration: str
        The duration of the video in the format 'HH:MM:SS'.
    language: str
        The language of the video.
    viewable: str
        Indicates if the video is viewable.
    stream_id: Optional[str]
        The ID of the stream from which the video was created.
    created_at: str
        The timestamp when the video was created.
    view_count: str
        The number of views the video has.
    description: str
        The description of the video.
    published_at: str
        The timestamp when the video was published.
    thumbnail_url: str
        The URL of the video's thumbnail.
    muted_segments: Optional[VideoMutedSegments]
        A list of muted segments in the video.
    """
    id: str
    url: str
    type: Literal['archive', 'highlight', 'upload']
    title: str
    duration: str
    language: str
    viewable: str
    stream_id: Optional[str]
    created_at: str
    view_count: str
    description: str
    published_at: str
    thumbnail_url: str
    muted_segments: Optional[VideoMutedSegments]


# Extension
class Extension(TypedDict):
    """
    Represents an extension.

    Attributes
    ----------
    id: str
        The ID of the extension.
    version: str
        The version of the extension.
    name: str
        The name of the extension.
    can_activate: bool
        Indicates if the extension can be activated.
    type: List[Literal['component', 'mobile', 'overlay', 'panel']]
        The types of the extension.
    """
    id: str
    version: str
    name: str
    can_activate: bool
    type: List[Literal['component', 'mobile', 'overlay', 'panel']]


class ExtensionDetails(TypedDict, total=False):
    """
    Represents details of an extension.

    Attributes
    ----------
    active: bool
        Indicates if the extension is active.
    id: NotRequired[str]
        The ID of the extension (optional).
    version: NotRequired[str]
        The version of the extension (optional).
    name: NotRequired[str]
        The name of the extension (optional).
    x: NotRequired[int]
        The x-coordinate for positioning the extension (optional).
    y: NotRequired[int]
        The y-coordinate for positioning the extension (optional).
    """
    active: bool
    id: NotRequired[str]
    version: NotRequired[str]
    name: NotRequired[str]
    x: NotRequired[int]
    y: NotRequired[int]


class ActiveExtensions(TypedDict):
    """
    Represents active extensions.

    Attributes
    ----------
    panel: Dict[str, ExtensionDetails]
        The active panel extensions.
    overlay: Dict[str, ExtensionDetails]
        The active overlay extensions.
    component: Dict[str, ExtensionDetails]
        The active component extensions.
    """
    panel: Dict[str, ExtensionDetails]
    overlay: Dict[str, ExtensionDetails]
    component: Dict[str, ExtensionDetails]


# Teams
class ChannelTeam(TypedDict):
    """
    Represents a channel team.

    Attributes
    ----------
    background_image_url: str
        The URL of the team's background image.
    banner: str
        The team's banner.
    created_at: str
        The timestamp when the team was created.
    updated_at: str
        The timestamp when the team was last updated.
    info: str
        Information about the team.
    thumbnail_url: str
        The URL of the team's thumbnail image.
    team_name: str
        The name of the team.
    team_display_name: str
        The display name of the team.
    id: str
        The ID of the team.
    """
    background_image_url: str
    banner: str
    created_at: str
    updated_at: str
    info: str
    thumbnail_url: str
    team_name: str
    team_display_name: str
    id: str


class Team(TypedDict):
    """
    Represents a team.

    Attributes
    ----------
    background_image_url: str
        The URL of the team's background image.
    banner: str
        The team's banner.
    created_at: str
        The timestamp when the team was created.
    updated_at: str
        The timestamp when the team was last updated.
    info: str
        Information about the team.
    thumbnail_url: str
        The URL of the team's thumbnail image.
    team_name: str
        The name of the team.
    team_display_name: str
        The display name of the team.
    id: str
        The ID of the team.
    users: List[SpecificUser]
        A list of users in the team.
    """
    background_image_url: str
    banner: str
    created_at: str
    updated_at: str
    info: str
    thumbnail_url: str
    team_name: str
    team_display_name: str
    id: str
    users: List[SpecificUser]



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster, SpecificUser
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import Literal, Optional, List


# Message
class ChatClearEvent(SpecificBroadcaster):
    """
    Represents an event where the chat is cleared.

    Attributes
    ----------
    (Inherits from SpecificBroadcaster)
    """
    pass


class ClearUserMessagesEvent(SpecificBroadcaster):
    """
    Represents an event where a user's messages are cleared.

    Attributes
    ----------
    target_user_id: str
        The ID of the target user whose messages are cleared.
    target_user_name: str
        The name of the target user whose messages are cleared.
    target_user_login: str
        The login name of the target user whose messages are cleared.
    """
    target_user_id: str
    target_user_name: str
    target_user_login: str


class Cheermote(TypedDict):
    """
    Represents a cheermote used in the chat.

    Attributes
    ----------
    prefix: str
        The prefix for the cheermote.
    bits: int
        The number of bits required to use the cheermote.
    tier: int
        The tier of the cheermote.
    """
    prefix: str
    bits: int
    tier: int


class Emote(TypedDict):
    """
    Represents an emote in the chat.

    Attributes
    ----------
    id: str
        The ID of the emote.
    emote_set_id: str
        The ID of the emote set to which this emote belongs.
    owner_id: str
        The ID of the owner of the emote.
    format: List[Literal['animated', 'static']]
        The formats in which the emote is available.
    """
    id: str
    emote_set_id: str
    owner_id: str
    format: List[Literal['animated', 'static']]


class Fragment(TypedDict):
    """
    Represents a fragment of a chat message.

    Attributes
    ----------
    type: Literal['text', 'cheermote', 'emote', 'mention']
        The type of the fragment.
    text: str
        The text content of the fragment.
    cheermote: Optional[Cheermote]
        The cheermote associated with the fragment, if applicable.
    emote: Optional[Emote]
        The emote associated with the fragment, if applicable.
    mention: Optional[SpecificUser]
        The user mentioned in the fragment, if applicable.
    """
    type: Literal['text', 'cheermote', 'emote', 'mention']
    text: str
    cheermote: Optional[Cheermote]
    emote: Optional[Emote]
    mention: Optional[SpecificUser]


class TextEmoteMessage(TypedDict):
    """
    Represents a chat message containing text and emotes.

    Attributes
    ----------
    text: str
        The text of the message.
    emotes: List[str]
        A list of emote IDs present in the message.
    """
    text: str
    emotes: List[str]


class Message(TypedDict):
    """
    Represents a chat message.

    Attributes
    ----------
    text: str
        The text content of the message.
    fragments: List[Fragment]
        A list of fragments that make up the message.
    """
    text: str
    fragments: List[Fragment]


class Badge(TypedDict):
    """
    Represents a badge displayed next to a user's name.

    Attributes
    ----------
    set_id: str
        The ID of the badge set to which this badge belongs.
    id: str
        The ID of the badge.
    info: str
        Additional information about the badge.
    """
    set_id: str
    id: str
    info: str


class Cheer(TypedDict):
    """
    Represents a cheer in the chat.

    Attributes
    ----------
    bits: int
        The number of bits used in the cheer.
    """
    bits: int


class Reply(TypedDict):
    """
    Represents a reply to a chat message.

    Attributes
    ----------
    parent_message_id: str
        The ID of the parent message being replied to.
    parent_message_body: str
        The body of the parent message.
    parent_user_id: str
        The ID of the user who sent the parent message.
    parent_user_name: str
        The name of the user who sent the parent message.
    parent_user_login: str
        The login name of the user who sent the parent message.
    thread_message_id: str
        The ID of the thread message in which the reply was made.
    thread_user_id: str
        The ID of the user who sent the thread message.
    thread_user_name: str
        The name of the user who sent the thread message.
    thread_user_login: str
        The login name of the user who sent the thread message.
    """
    parent_message_id: str
    parent_message_body: str
    parent_user_id: str
    parent_user_name: str
    parent_user_login: str
    thread_message_id: str
    thread_user_id: str
    thread_user_name: str
    thread_user_login: str


class MessageEvent(SpecificBroadcaster):
    """
    Represents a chat message event.

    Attributes
    ----------
    chatter_user_id: str
        The ID of the user who sent the message.
    chatter_user_name: str
        The name of the user who sent the message.
    chatter_user_login: str
        The login name of the user who sent the message.
    message_id: str
        A UUID that identifies the message.
    message: Message
        The structured chat message content.
        - text: str
            The plain text of the message.
        - fragments: List[Fragment]
            An ordered list of message fragments (e.g., text, cheermotes, emotes, mentions).
    message_type: str
        The type of the message. Possible values include:

        - text
        - channel_points_highlighted
        - channel_points_sub_only
        - user_intro
        - power_ups_message_effect
        - power_ups_gigantified_emote
    badges: List[Badge]
        A list of badges associated with the user.
    cheer: Optional[Cheer]
        Metadata if the message includes a cheer.
    color: Optional[str]
        The color of the user's chat message, in hexadecimal RGB format (e.g., "#FF0000").
    reply: Optional[Reply]
        Details of a reply to a message, if applicable.
    channel_points_custom_reward_id: Optional[str]
        The ID of the custom reward associated with the message, if applicable.
    channel_points_animation_id: Optional[str]
        The ID of the animation associated with the channel points reward, if applicable.
    source_broadcaster_user_id: Optional[str]
        The broadcaster user ID of the channel the message was sent from,
        in case of a shared chat session.
    source_broadcaster_user_name: Optional[str]
        The user name of the broadcaster of the channel the message was sent from,
        in case of a shared chat session.
    source_broadcaster_user_login: Optional[str]
        The login of the broadcaster of the channel the message was sent from,
        in case of a shared chat session.
    source_message_id: Optional[str]
        The UUID that identifies the source message from the channel it was sent from,
        in case of a shared chat session.
    source_badges: Optional[List[Badge]]
        The list of chat badges for the chatter in the channel the message was sent from,
        in case of a shared chat session.
    """
    chatter_user_id: str
    chatter_user_name: str
    chatter_user_login: str
    message_id: str
    message: Message
    message_type: Literal[
        'text',
        'channel_points_highlighted',
        'channel_points_sub_only',
        'user_intro',
        'power_ups_message_effect',
        'power_ups_gigantified_emote'
    ]
    badges: List[Badge]
    cheer: Optional[Cheer]
    color: Optional[str]
    reply: Optional[Reply]
    channel_points_custom_reward_id: Optional[str]
    channel_points_animation_id: Optional[str]
    source_broadcaster_user_id: Optional[str]
    source_broadcaster_user_name: Optional[str]
    source_broadcaster_user_login: Optional[str]
    source_message_id: Optional[str]
    source_badges: Optional[List[Badge]]


class MessageDeleteEvent(SpecificBroadcaster):
    """
    Represents an event where a chat message is deleted.

    Attributes
    ----------
    target_user_id: str
        The ID of the user whose message was deleted.
    target_user_name: str
        The name of the user whose message was deleted.
    target_user_login: str
        The login name of the user whose message was deleted.
    message_id: str
        The ID of the deleted message.
    """
    target_user_id: str
    target_user_name: str
    target_user_login: str
    message_id: str


class Sub(TypedDict):
    """
    Represents a subscription.

    Attributes
    ----------
    sub_tier: str
        The tier of the subscription.
    is_prime: bool
        Whether the subscription is a Prime subscription.
    duration_months: int
        The duration of the subscription in months.
    """
    sub_tier: str
    is_prime: bool
    duration_months: int


# Notification
class Resub(TypedDict):
    """
    Represents a resubscription.

    Attributes
    ----------
    cumulative_months: int
        The total number of months the user has subscribed.
    duration_months: int
        The duration of the resubscription in months.
    streak_months: int
        The number of months in the current streak.
    sub_tier: str
        The tier of the resubscription.
    is_prime: Optional[bool]
        Whether the resubscription is a Prime subscription, if applicable.
    is_gift: bool
        Whether the resubscription is a gift.
    gifter_is_anonymous: Optional[bool]
        Whether the gifter is anonymous, if applicable.
    gifter_user_id: Optional[str]
        The ID of the user who gifted the subscription, if applicable.
    gifter_user_name: Optional[str]
        The name of the user who gifted the subscription, if applicable.
    gifter_user_login: Optional[str]
        The login name of the user who gifted the subscription, if applicable.
    """
    cumulative_months: int
    duration_months: int
    streak_months: int
    sub_tier: str
    is_prime: Optional[bool]
    is_gift: bool
    gifter_is_anonymous: Optional[bool]
    gifter_user_id: Optional[str]
    gifter_user_name: Optional[str]
    gifter_user_login: Optional[str]


class SubGift(TypedDict):
    """
    Represents a subscription gift.

    Attributes
    ----------
    duration_months: int
        The duration of the gift subscription in months.
    cumulative_total: Optional[int]
        The total number of gifted subscriptions, if applicable.
    recipient_user_id: str
        The ID of the recipient of the gift subscription.
    recipient_user_name: str
        The name of the recipient of the gift subscription.
    recipient_user_login: str
        The login name of the recipient of the gift subscription.
    sub_tier: str
        The tier of the gift subscription.
    community_gift_id: Optional[str]
        The ID of the community gift, if applicable.
    """
    duration_months: int
    cumulative_total: Optional[int]
    recipient_user_id: str
    recipient_user_name: str
    recipient_user_login: str
    sub_tier: str
    community_gift_id: Optional[str]


class CommunitySubGift(TypedDict):
    """
    Represents a community subscription gift.

    Attributes
    ----------
    id: str
        The ID of the community gift.
    total: int
        The total number of subscriptions gifted to the community.
    sub_tier: str
        The tier of the gift subscriptions.
    cumulative_total: Optional[int]
        The cumulative total number of gifted subscriptions, if applicable.
    """
    id: str
    total: int
    sub_tier: str
    cumulative_total: Optional[int]


class GiftPaidUpgrade(TypedDict):
    """
    Represents a paid upgrade for a gifted subscription.

    Attributes
    ----------
    gifter_is_anonymous: bool
        Whether the gifter is anonymous.
    gifter_user_id: Optional[str]
        The ID of the user who gifted the upgrade, if applicable.
    gifter_user_name: Optional[str]
        The name of the user who gifted the upgrade, if applicable.
    gifter_user_login: Optional[str]
        The login name of the user who gifted the upgrade, if applicable.
    """
    gifter_is_anonymous: bool
    gifter_user_id: Optional[str]
    gifter_user_name: Optional[str]
    gifter_user_login: Optional[str]


class PrimePaidUpgrade(TypedDict):
    """
    Represents a paid upgrade for a Prime subscription.

    Attributes
    ----------
    sub_tier: str
        The tier of the upgraded Prime subscription.
    """
    sub_tier: str


class Raid(TypedDict):
    """
    Represents a raid event.

    Attributes
    ----------
    user_id: str
        The ID of the user who initiated the raid.
    user_name: str
        The name of the user who initiated the raid.
    user_login: str
        The login name of the user who initiated the raid.
    viewer_count: int
        The number of viewers participating in the raid.
    profile_image_url: str
        The URL of the user's profile image.
    """
    user_id: str
    user_name: str
    user_login: str
    viewer_count: int
    profile_image_url: str


class PayItForward(TypedDict):
    """
    Represents a Pay It Forward event.

    Attributes
    ----------
    gifter_is_anonymous: bool
        Whether the gifter is anonymous.
    gifter_user_id: Optional[str]
        The ID of the user who initiated the Pay It Forward, if applicable.
    gifter_user_name: Optional[str]
        The name of the user who initiated the Pay It Forward, if applicable.
    gifter_user_login: Optional[str]
        The login name of the user who initiated the Pay It Forward, if applicable.
    """
    gifter_is_anonymous: bool
    gifter_user_id: Optional[str]
    gifter_user_name: Optional[str]
    gifter_user_login: Optional[str]


class Announcement(TypedDict):
    """
    Represents an announcement event.

    Attributes
    ----------
    color: str
        The color of the announcement.
    """
    color: str


class CharityDonationAmount(TypedDict):
    """
    Represents the amount donated to charity.

    Attributes
    ----------
    value: int
        The donation amount.
    decimal_place: int
        The number of decimal places in the donation amount.
    currency: str
        The currency of the donation.
    """
    value: int
    decimal_place: int
    currency: str


class CharityDonation(TypedDict):
    """
    Represents a charity donation event.

    Attributes
    ----------
    charity_name: str
        The name of the charity receiving the donation.
    amount: CharityDonationAmount
        The amount of the donation.
    """
    charity_name: str
    amount: CharityDonationAmount


class BitsBadgeTier(TypedDict):
    """
    Represents a bits badge tier.

    Attributes
    ----------
    tier: int
        The tier of the bits badge.
    """
    tier: int


class NotificationEvent(SpecificBroadcaster):
    """
    Represents a notification event.

    Attributes
    ----------
    broadcaster_user_id: str
        ID of the broadcaster.
    broadcaster_user_name: str
        Name of the broadcaster.
    broadcaster_user_login: str
        Login of the broadcaster.
    chatter_user_id: str
        ID of the chatter.
    chatter_user_name: str
        Name of the chatter.
    chatter_user_login: str
        Login of the chatter.
    chatter_is_anonymous: bool
        Whether the chatter is anonymous.
    color: str
        Color of the notification.
    badges: List[Badge]
        Badges associated with the chatter.
    system_message: Optional[str]
        System message of the notification.
    message_id: str
        ID of the message.
    message: Message
        Content of the message.
    notice_type: str
        Type of the notification.
    sub: Optional[Sub]
        Subscription details, if applicable.
    resub: Optional[Resub]
        Resubscription details, if applicable.
    sub_gift: Optional[SubGift]
        Subscription gift details, if applicable.
    community_sub_gift: Optional[CommunitySubGift]
        Community subscription gift details, if applicable.
    gift_paid_upgrade: Optional[GiftPaidUpgrade]
        Paid upgrade for a gifted subscription, if applicable.
    prime_paid_upgrade: Optional[PrimePaidUpgrade]
        Paid upgrade for a Prime subscription, if applicable.
    raid: Optional[Raid]
        Raid event details, if applicable.
    unraid: Optional[SpecificUser]
        Unraid event details, if applicable.
    pay_it_forward: Optional[PayItForward]
        Pay It Forward event details, if applicable.
    announcement: Optional[Announcement]
        Announcement details, if applicable.
    charity_donation: Optional[CharityDonation]
        Charity donation details, if applicable.
    bits_badge_tier: Optional[BitsBadgeTier]
        Bits badge tier details, if applicable.
    source_broadcaster_user_id: Optional[str]
        ID of the source broadcaster, if different.
    source_broadcaster_user_name: Optional[str]
        Name of the source broadcaster, if different.
    source_broadcaster_user_login: Optional[str]
        Login of the source broadcaster, if different.
    source_message_id: Optional[str]
        ID of the source message, if different.
    source_badges: Optional[List[Badge]]
        Badges for the source chatter, if different.
    shared_chat_sub: Optional[Sub]
        Shared chat subscription details, if applicable.
    shared_chat_resub: Optional[Resub]
        Shared chat re-subscription details, if applicable.
    shared_chat_sub_gift: Optional[SubGift]
        Shared chat subscription gift details, if applicable.
    shared_chat_community_sub_gift: Optional[CommunitySubGift]
        Shared chat community subscription gift details, if applicable.
    shared_chat_gift_paid_upgrade: Optional[GiftPaidUpgrade]
        Shared chat paid upgrade for a gifted subscription, if applicable.
    shared_chat_prime_paid_upgrade: Optional[PrimePaidUpgrade]
        Shared chat paid upgrade for a Prime subscription, if applicable.
    shared_chat_pay_it_forward: Optional[PayItForward]
        Shared chat Pay It Forward event details, if applicable.
    shared_chat_raid: Optional[Raid]
        Shared chat raid event details, if applicable.
    shared_chat_announcement: Optional[Announcement]
        Shared chat announcement details, if applicable.
    """
    tier: int
    chatter_user_id: str
    chatter_user_name: str
    chatter_user_login: str
    chatter_is_anonymous: bool
    color: str
    badges: List[Badge]
    system_message: Optional[str]
    message_id: str
    message: Message
    notice_type: str
    sub: Optional[Sub]
    resub: Optional[Resub]
    sub_gift: Optional[SubGift]
    community_sub_gift: Optional[CommunitySubGift]
    gift_paid_upgrade: Optional[GiftPaidUpgrade]
    prime_paid_upgrade: Optional[PrimePaidUpgrade]
    pay_it_forward: Optional[PayItForward]
    raid: Optional[Raid]
    unraid: Optional[SpecificUser]
    announcement: Optional[Announcement]
    bits_badge_tier: Optional[BitsBadgeTier]
    charity_donation: Optional[CharityDonation]
    source_broadcaster_user_id: Optional[str]
    source_broadcaster_user_name: Optional[str]
    source_broadcaster_user_login: Optional[str]
    source_message_id: Optional[str]
    source_badges: Optional[List[Badge]]
    shared_chat_sub: Optional[Sub]
    shared_chat_resub: Optional[Resub]
    shared_chat_sub_gift: Optional[SubGift]
    shared_chat_community_sub_gift: Optional[CommunitySubGift]
    shared_chat_gift_paid_upgrade: Optional[GiftPaidUpgrade]
    shared_chat_prime_paid_upgrade: Optional[PrimePaidUpgrade]
    shared_chat_pay_it_forward: Optional[PayItForward]
    shared_chat_raid: Optional[Raid]
    shared_chat_announcement: Optional[Announcement]


# Settings
class SettingsUpdateEvent(SpecificBroadcaster):
    """
    Represents an event where chat settings are updated.

    Attributes
    ----------
    emote_mode: bool
        Whether emote mode is enabled.
    follower_mode: bool
        Whether follower mode is enabled.
    follower_mode_duration_minutes: Optional[int]
        The duration of follower mode in minutes, if applicable.
    slow_mode: bool
        Whether slow mode is enabled.
    slow_mode_wait_time_seconds: Optional[int]
        The wait time for slow mode in seconds, if applicable.
    subscriber_mode: bool
        Whether subscriber mode is enabled.
    unique_chat_mode: bool
        Whether unique chat mode is enabled.
    """
    emote_mode: bool
    follower_mode: bool
    follower_mode_duration_minutes: Optional[int]
    slow_mode: bool
    slow_mode_wait_time_seconds: Optional[int]
    subscriber_mode: bool
    unique_chat_mode: bool


class SharedChatBeginEvent(SpecificBroadcaster):
    """
    Represents the beginning of a channel shared chat session event.

    Attributes
    ----------
    session_id: str
        The unique identifier for the shared chat session.
    host_broadcaster_user_id: str
        The User ID of the host channel.
    host_broadcaster_user_name: str
        The display name of the host channel.
    host_broadcaster_user_login: str
        The user login of the host channel.
    participants: List[SpecificBroadcaster]
        The list of participants in the session.
    """
    session_id: str
    host_broadcaster_user_id: str
    host_broadcaster_user_name: str
    host_broadcaster_user_login: str
    participants: List[SpecificBroadcaster]


class SharedChatUpdateEvent(SpecificBroadcaster):
    """
    Represents an update to a channel shared chat session event.

    Attributes
    ----------
    session_id: str
        The unique identifier for the shared chat session.
    host_broadcaster_user_id: str
        The User ID of the host channel.
    host_broadcaster_user_name: str
        The display name of the host channel.
    host_broadcaster_user_login: str
        The user login of the host channel.
    participants: List[SpecificBroadcaster]
        The list of participants in the session.
    """
    session_id: str
    host_broadcaster_user_id: str
    host_broadcaster_user_name: str
    host_broadcaster_user_login: str
    participants: List[SpecificBroadcaster]


class SharedChatEndEvent(SpecificBroadcaster):
    """
    Represents the end of a channel shared chat session event.

    Attributes
    ----------
    session_id: str
        The unique identifier for the shared chat session.
    host_broadcaster_user_id: str
        The User ID of the host channel.
    host_broadcaster_user_name: str
        The display name of the host channel.
    host_broadcaster_user_login: str
        The user login of the host channel.
    """
    session_id: str
    host_broadcaster_user_id: str
    host_broadcaster_user_name: str
    host_broadcaster_user_login: str



"""
twitch.ext.overlay

An extension that provides functionalities to manage and send overlay alerts,
handle WebSocket connections, and manage file attachments.

:copyright: (c) 2024-present Snifo
:license: MIT, see LICENSE for more details.
"""


from .overlay import Overlay



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import Broadcaster, SpecificUser
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import List, Literal, Optional


class Entitlement(TypedDict):
    """
    Represents an entitlement for a user.

    Attributes
    ----------
    id: str
        The unique identifier of the entitlement.
    benefit_id: str
        The ID of the benefit associated with the entitlement.
    timestamp: str
        The timestamp when the entitlement was created.
    user_id: str
        The ID of the user who owns the entitlement.
    game_id: str
        The ID of the game associated with the entitlement.
    fulfillment_status: Literal['CLAIMED', 'FULFILLED']
        The status of the entitlement, indicating whether it has been claimed or fulfilled.
    last_updated: str
        The timestamp when the entitlement was last updated.
    """
    id: str
    benefit_id: str
    timestamp: str
    user_id: str
    game_id: str
    fulfillment_status: Literal['CLAIMED', 'FULFILLED']
    last_updated: str


class EntitlementsUpdate(TypedDict):
    """
    Represents an update to multiple entitlements.

    Attributes
    ----------
    status: str
        The status of the entitlement update process.
    ids: List[str]
        A list of entitlement IDs that have been updated.
    """
    status: str
    ids: List[str]


class CharityAmount(TypedDict):
    """
    Represents an amount in a charity context.

    Attributes
    ----------
    value: int
        The value of the amount.
    decimal_places: int
        The number of decimal places in the value.
    currency: str
        The currency of the amount, e.g., 'USD', 'EUR'.
    """
    value: int
    decimal_places: int
    currency: str


class Charity(Broadcaster):
    """
    Represents a charity campaign.

    Attributes
    ----------
    id: str
        The unique identifier of the charity campaign.
    charity_logo: str
        The URL to the logo of the charity.
    charity_name: str
        The name of the charity.
    target_amount: Optional[CharityAmount]
        The target amount for the charity campaign.
    current_amount: CharityAmount
        The current amount raised by the charity campaign.
    charity_website: str
        The website URL of the charity.
    charity_description: str
        A description of the charity.
    """
    id: str
    charity_logo: str
    charity_name: str
    target_amount: Optional[CharityAmount]
    current_amount: CharityAmount
    charity_website: str
    charity_description: str


class CharityDonation(SpecificUser):
    """
    Represents a donation to a charity campaign.

    Attributes
    ----------
    id: str
        The unique identifier of the donation.
    amount: CharityAmount
        The amount donated.
    campaign_id: str
        The ID of the charity campaign to which the donation was made.
    """
    id: str
    amount: CharityAmount
    campaign_id: str


class Goal(Broadcaster):
    """
    Represents a goal in a Twitch context.

    Attributes
    ----------
    id: int
        The unique identifier of the goal.
    type: str
        The type of the goal, e.g., 'follower', 'subscriber'.
    description: str
        A brief description of the goal.
    target_amount: int
        The target amount to achieve the goal.
    current_amount: int
        The current progress towards the goal.
    created_at: str
        The timestamp when the goal was created.
    """
    id: int
    type: str
    description: str
    target_amount: int
    current_amount: int
    created_at: str



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from twitch import utils
from typing import TYPE_CHECKING
from .http import HTTPClient
from .enums import Scopes
import asyncio

if TYPE_CHECKING:
    from typing import Optional, Type, Self, List, Union, Tuple
    from types import TracebackType
    from twitch.types import users
    from twitch import Client

import logging
_logger = logging.getLogger(__name__)

__all__ = ('DeviceAuthFlow',)


class DeviceAuthFlow:
    """
    A class to handle the device authentication flow.

    Parameters
    ----------
    client: Client
        The Twitch client instance.
    scopes: List[Union[str, Scope]]
        List of scopes for the authentication.
    dispatch: bool
        If True, Will automatically trigger relevant events such as `on_code` and `on_auth`.
    wrap_run: bool
        If True, automatically assigns the `run` method to the client's `run` attribute.

    Example
    -------
    Example usage when wrap_run is True:

    ```python
    from twitch import Client
    from twitch.ext.oauth import DeviceAuthFlow

    client = Client(client_id=..., client_secret=....)
    # Adds support for authentication.
    DeviceAuthFlow(client=client, scopes=['user:read:email'])

    @client.event
    async def on_code(code: str):
        print(f' verification uri: https://www.twitch.tv/activate?device-code={code}')

    # ........ do weird stuff here *-*.

    # Must be called at the end.
    client.run(l)
    ```
    """

    __slots__ = ('_client', '_wrap_run', '_http', 'scopes', '_dispatch')

    def __init__(self,
                 client: Client,
                 *,
                 scopes: List[Union[str, Scopes]],
                 dispatch: bool = True,
                 wrap_run: bool = True
                 ) -> None:
        self._client: Client = client
        self._dispatch: bool = dispatch
        self._wrap_run: bool = wrap_run
        self._http: HTTPClient = HTTPClient(self._client.client_id)
        if self._wrap_run:
            self._client.run = self._run  # type: ignore
        self.scopes: List[str] = [scope.value if isinstance(scope, Scopes) else scope for scope in scopes]

    async def __aenter__(self) -> Self:
        """
        Enter the runtime context related to this object.
        """
        return self

    async def __aexit__(self, exc_type: Optional[Type[BaseException]],
                        exc_value: Optional[BaseException],
                        traceback: Optional[TracebackType]) -> None:
        """
        Exit the runtime context related to this object.
        """
        await self.close()

    def initiate_event_loop(self):
        if self._http.loop is not None:
            return

        if self._client.loop is not None:
            loop = self._client.loop
        else:
            loop = asyncio.get_running_loop()

        self._http.loop = loop
        self._client.loop = loop
        self._client.http.loop = loop

    async def get_device_code(self) -> Tuple[str, str, int, int]:
        """
        Initiates the device authorization flow and retrieves the device code.

        Returns
        -------
        Tuple
            A tuple containing the user code, device code, expiration time, and polling interval.
        """
        self.initiate_event_loop()
        data: users.DeviceAuthFlow = await self._http.start_device_auth_flow(scopes=self.scopes)
        if self._dispatch:
            self._client.dispatch('code', data['user_code'])
        return data['user_code'], data['device_code'], data['expires_in'], data['interval']

    async def revoke_token(self, token: str) -> None:
        """
        Revoke a token.

        Parameters
        ----------
        token: str
            The token to revoke (access token or refresh token).
        """
        self.initiate_event_loop()
        await self._http.revoke_token(token=token)

    async def poll_for_authorization(self, device_code: str, expires_in: int, interval: int) -> tuple[str, str]:
        """
        Polls the Twitch API to check if the user has authorized the application.

        Parameters
        ----------
        device_code: str
            The device code received from the get_device_code method.
        expires_in: int
            The expiration time of the device code.
        interval: int
            The polling interval.

        Returns
        -------
        tuple
            A tuple containing the access token and refresh token.
        """
        self.initiate_event_loop()
        data: users.OAuthRefreshToken = await self._http.poll_for_token(device_code,
                                                                        expires_in=expires_in,
                                                                        interval=interval)
        if self._dispatch:
            self._client.dispatch('auth', data['access_token'], data['refresh_token'])
        return data['access_token'], data['refresh_token']

    async def close(self):
        """
        Closes the HTTP client session.
        """
        await self._http.close()

    def _run(self,
             *,
             reconnect: bool = True,
             log_handler: Optional[logging.Handler] = None,
             log_level: Optional[int] = None,
             root_logger: bool = False):
        """
        Run the client with the authentication flow and handle the event loop.

        !!! danger
            This function must be the last function to call as it blocks
            the execution of anything after it.

        Parameters
        ----------
        reconnect: Optional[str]
            Whether to attempt reconnection.
        log_handler: Optional[logging.Handler]
            The logging handler.
        log_level: Optional[int]
            The logging level. Defaults to MISSING, which gets replaced with logging.INFO.
        root_logger: bool
            If True, configure the root logger. If False, configure a logger specific to the calling module.
            Defaults to False.
        """
        if log_handler is None:
            utils.setup_logging(handler=log_handler, level=log_level, root=root_logger)

        async def runner():
            async with self:
                user_code, device_code, expires_in, interval = await self.get_device_code()
                access_token, refresh_token = await self.poll_for_authorization(device_code, expires_in, interval)

            async with self._client:
                await self._client.start(access_token, refresh_token, reconnect=reconnect)

        try:
            asyncio.run(runner())
        except KeyboardInterrupt:
            return



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster, SpecificUser
from .chat import TextEmoteMessage
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Optional


# Update
class ChannelUpdateEvent(SpecificBroadcaster):
    """
    Represents an event where the broadcaster updates their channel information.

    Attributes
    ----------
    title: str
        The title of the stream.
    language: str
        The language of the stream.
    category_id: str
        The ID of the category.
    category_name: str
        The name of the category.
    content_classification_labels: List[str]
        A list of content classification labels associated with the stream.
    """
    title: str
    language: str
    category_id: str
    category_name: str
    content_classification_labels: List[str]


# Follow
class FollowEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user follows a broadcaster.

    Attributes
    ----------
    followed_at: str
        The timestamp when the follow occurred.
    """
    followed_at: str


# Subscription
class SubscribeEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user subscribes to a broadcaster's channel.

    Attributes
    ----------
    tier: str
        The tier of the subscription (e.g., Tier 1, Tier 2, Tier 3).
    is_gift: bool
        Indicates whether the subscription was gifted.
    """
    tier: str
    is_gift: bool


class SubscriptionEndEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user's subscription to a broadcaster's channel ends.

    Attributes
    ----------
    tier: str
        The tier of the subscription that ended.
    is_gift: bool
        Indicates whether the ended subscription was a gift.
    """
    tier: str
    is_gift: bool


class SubscriptionGiftEvent(SpecificBroadcaster):
    """
    Represents an event where a user gifts a subscription to another user.

    Attributes
    ----------
    user_id: Optional[str]
        The ID of the user who gifted the subscription, if available.
    user_login: Optional[str]
        The login name of the user who gifted the subscription, if available.
    user_name: Optional[str]
        The display name of the user who gifted the subscription, if available.
    total: int
        The total number of subscriptions gifted by the user.
    tier: str
        The tier of the gifted subscription (e.g., Tier 1, Tier 2, Tier 3).
    cumulative_total: Optional[int]
        The cumulative total number of subscriptions gifted by the user, if available.
    is_anonymous: bool
        Indicates whether the gift was given anonymously.
    """
    user_id: Optional[str]
    user_login: Optional[str]
    user_name: Optional[str]
    total: int
    tier: str
    cumulative_total: Optional[int]
    is_anonymous: bool


class SubscriptionMessageEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user sends a message during their subscription.

    Attributes
    ----------
    tier: str
        The tier of the subscription associated with the message.
    message: TextEmoteMessage
        The message content, which may include text and emotes.
    cumulative_months: int
        The total number of months the user has been subscribed.
    streak_months: Optional[int]
        The number of consecutive months the user has been subscribed, if available.
    duration_months: int
        The duration of the current subscription in months.
    """
    tier: str
    message: TextEmoteMessage
    cumulative_months: int
    streak_months: Optional[int]
    duration_months: int


# VIP Add/Remove
class VIPAddEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user is granted VIP status in a broadcaster's channel.
    """
    pass


class VIPRemoveEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user is removed from VIP status in a broadcaster's channel.
    """
    pass



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .errors import ConnectionClosed
from typing import TYPE_CHECKING
from json import JSONDecodeError
import aiohttp
import asyncio
import json

if TYPE_CHECKING:
    from typing import Optional, Set, ClassVar, Any, Self
    from .state import ConnectionState
    from .client import Client


import logging
_logger = logging.getLogger(__name__)

__all__ = ('EventSubWebSocket', 'ReconnectWebSocket')


class WebSocketClosure(Exception):
    """Exception indicating closure of the WebSocket."""


class ReconnectWebSocket(Exception):
    """Exception indicating the need to reconnect to the websocket."""

    def __init__(self, url: str) -> None:
        self.url: str = url


class EventSubWebSocket:
    """Represents Twitch EventSub WebSocket."""
    DEFAULT_GATEWAY: ClassVar[str] = 'wss://eventsub.wss.twitch.tv/ws?keepalive_timeout_seconds=30'
    SESSION_WELCOME: ClassVar[str] = 'session_welcome'
    SESSION_KEEPALIVE: ClassVar[str] = 'session_keepalive'
    SESSION_RECONNECT: ClassVar[str] = 'session_reconnect'
    NOTIFICATION: ClassVar[str] = 'notification'
    REVOCATION: ClassVar[str] = 'revocation'

    def __init__(self,
                 socket: aiohttp.ClientWebSocketResponse,
                 state: ConnectionState,
                 *,
                 loop: asyncio.AbstractEventLoop) -> None:
        self.socket: aiohttp.ClientWebSocketResponse = socket
        self.loop: asyncio.AbstractEventLoop = loop
        self.session_id: Optional[str] = None
        self._subscriptions_task: Optional[asyncio.Task] = None
        self._state: ConnectionState = state
        self._timeout: Optional[int] = None

    async def close(self, code: int = 3000) -> None:
        """Close the WebSocket connection."""
        if self._subscriptions_task is not None and not self._subscriptions_task.done():
            self._subscriptions_task.cancel()
        await self.socket.close(code=code)

    @property
    def is_open(self) -> bool:
        return not self.socket.closed

    async def create_subscriptions(self, *, events: Set[str], initial: bool) -> None:
        """initial subscriptions."""
        if initial:
            for event in events:
                try:
                    await self._state.create_subscription(self._state.user.id, event, self.session_id)
                except Exception as exc:
                    _logger.exception('Error processing event `on_%s`: %s', event, str(exc))
        else:
            await self._state.initialize_after_disconnect(self.session_id)
        self._state.state_ready()

    @classmethod
    async def initialize_websocket(cls,
                                   client: Client,
                                   state: ConnectionState,
                                   gateway: Optional[str] = None,
                                   resume: bool = False,
                                   initial: bool = False) -> EventSubWebSocket:
        """Initialize websocket connection."""
        gateway = gateway or cls.DEFAULT_GATEWAY
        socket = await client.http.ws_connect(url=gateway, resume=resume)
        state.ws_connect()
        ws: Self = cls(socket, state, loop=client.loop)
        # Waits for welcome message.
        await ws.poll_handle_dispatch()
        if not resume:
            # Get default events.
            events = {attr.replace('on_', '', 1) for attr in dir(client) if attr.startswith('on_')}
            # Add additional default events.
            events.update({'channel_update', 'user_update', 'stream_online', 'stream_offline'})
            task = ws.create_subscriptions(events=events, initial=initial)
            ws._subscriptions_task = ws.loop.create_task(task, name='twitch:gateway:subscriptions')
        else:
            state.state_ready()
        return ws

    async def poll_handle_dispatch(self) -> None:
        """Polls for a DISPATCH event and manages the gateway loop."""
        try:
            # 5 seconds incase something went wrong.
            msg = await self.socket.receive(timeout=(self._timeout or 30) + 5)
            if msg.type is aiohttp.WSMsgType.TEXT:
                await self.received_message(data=msg.data)
            elif msg.type is aiohttp.WSMsgType.ERROR:
                _logger.debug('Received error %s', msg)
                raise WebSocketClosure
            elif msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSE):
                _logger.debug('Received %s', msg)
                raise WebSocketClosure
        except (asyncio.TimeoutError, WebSocketClosure) as exc:
            code = self.socket.close_code
            raise ConnectionClosed(self.socket, code=code) from exc

    async def received_message(self, *, data: Any) -> None:
        """Handle received message."""
        try:
            data = json.loads(data)
            await self._state.socket_raw_receive(data=data)
        except (UnicodeDecodeError, JSONDecodeError) as error:
            _logger.exception('Failed to parse response as JSON: %s. Response: %s', error)
            return
        if data.get('metadata'):
            metadata = data['metadata']
            if metadata['message_type'] == self.SESSION_WELCOME:
                _session = data['payload']['session']
                _logger.debug('Connected to WebSocket. Session ID: %s', _session['id'])
                self.session_id = _session['id']
                self._timeout = _session['keepalive_timeout_seconds']
                return
            if metadata['message_type'] == self.SESSION_KEEPALIVE:
                _logger.debug('Received a keepalive message. The WebSocket connection is healthy.')
                return

            if metadata['message_type'] == self.SESSION_RECONNECT:
                await self.close()
                self._state.ws_disconnect()
                raise ReconnectWebSocket(url=data['payload']['session']['reconnect_url'])

            if metadata['message_type'] == self.NOTIFICATION:
                self._state.parse(data=data['payload'])
                return

            if metadata['message_type'] == self.REVOCATION:
                _logger.warning(
                    'Subscription Revoked: ID %s, Type "%s", Status "%s"',
                    data['payload']['subscription']['id'],
                    data['payload']['subscription']['type'],
                    data['payload']['subscription']['status'],
                )
                return



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Literal, Dict


class Broadcaster(TypedDict):
    """
    Represents a broadcaster.

    Attributes
    ----------
    broadcaster_id: str
        The unique identifier for the broadcaster.
    broadcaster_name: str
        The display name of the broadcaster.
    broadcaster_login: str
        The login name of the broadcaster.
    """
    broadcaster_id: str
    broadcaster_name: str
    broadcaster_login: str


class Moderator(TypedDict):
    """
    Represents a moderator.

    Attributes
    ----------
    moderator_id: str
        The unique identifier for the moderator.
    moderator_name: str
        The display name of the moderator.
    moderator_login: str
        The login name of the moderator.
    """
    moderator_id: str
    moderator_name: str
    moderator_login: str


class SpecificUser(TypedDict):
    """
    Represents a specific user.

    Attributes
    ----------
    user_id: str
        The unique identifier for the user.
    user_name: str
        The display name of the user.
    user_login: str
        The login name of the user.
    """
    user_id: str
    user_name: str
    user_login: str


# Auth
class OAuthToken(TypedDict):
    client_id: str
    login: str
    scopes: List[str]
    user_id: str
    expires_in: int


class OAuthRefreshToken(TypedDict):
    access_token: str
    expires_in: int
    refresh_token: str
    token_type: str


class DeviceAuthFlow(TypedDict):
    device_code: str
    expires_in: int
    interval: int
    user_code: str
    verification_uri: str


class Transport(TypedDict):
    method: Literal['webhook', 'websocket', 'conduit']
    callback: str


class EventSubSubscription(TypedDict):
    id: str
    status: Literal['webhook_callback_verification_pending', 'enabled', 'disabled']
    type: str
    version: str
    condition: Dict[str, str]
    created_at: str
    transport: Transport
    cost: int


# User

class User(TypedDict):
    """
    Represents a user.

    Attributes
    ----------
    id: str
        The unique identifier for the user.
    login: str
        The login name of the user.
    display_name: str
        The display name of the user.
    type: Literal['admin', 'global_mod', 'staff', '']
        The type of user.
    broadcaster_type: Literal['affiliate', 'partner', '']
        The broadcaster type of the user.
    description: str
        A brief description of the user.
    profile_image_url: str
        The URL to the user's profile image.
    offline_image_url: str
        The URL to the user's offline image.
    email: str
        The email address of the user.
    created_at: str
        The date and time when the user account was created, in ISO 8601 format.
    """
    id: str
    login: str
    display_name: str
    type: Literal['admin', 'global_mod', 'staff', '']
    broadcaster_type: Literal['affiliate', 'partner', '']
    description: str
    profile_image_url: str
    offline_image_url: str
    email: str
    created_at: str
